import { IDataBind, IDataSourceQuery, deserializeDynamicMap, EComponentType } from '@cloudbase/cals';
import { IReactionDisposer, autorun, observable, toJS } from 'mobx';
import {
  IGenerateContext,
  IGenerateOptions,
  generateDataBinds,
  generateListnerInstances,
  splitDynamicMap,
} from './adapter/generate';
import { generateEmit, mergeDynamic2StaticData } from './runtime';
import { debounce } from 'lodash-es';

export interface IQueryContext extends IGenerateContext {
  $w: any;
}

export class Query {
  private _schema: IDataSourceQuery;
  private _context: IQueryContext;
  private _disposes: IReactionDisposer[] = [];
  private _dataBinds: Record<string, Function> = {};
  private _triggered = false;
  private _action: any;
  private _paramsRef: { current: any } = observable({ current: null });
  private _emit: (eventName, data) => Promise<any>;
  private _currentRequestKey = null;
  private _observableValue: { data: any; error: Error } = observable({ data: null, error: null });

  constructor({
    schema,
    context,
    options,
  }: {
    schema: IDataSourceQuery;
    context: IQueryContext;
    options?: IGenerateOptions;
  }) {
    const { looseError = false, useScopeWidgetData = false } = options;
    const { $w, $app, $page } = context;
    this._schema = schema;
    this._context = context;

    if (this._schema?.trigger === 'auto') {
      this._disposes.push(
        autorun(
          (r) => {
            try {
              const data = this._resolveParams();
              if (this._triggered) {
                this._debounceTrigger(data);
              }
            } catch (e) {
              console.error(e);
            }
          },
          { delay: 50 },
        ),
      );
    }

    const baseParam: {
      dataSourceName: string;
      sqlTemplateId?: string;
      methodName?: string;
    } = {
      dataSourceName: this._schema.dataSourceName,
    };

    if (this._schema?.type === 'sql') {
      baseParam.sqlTemplateId = this._schema.method.sqlTemplateId;
    } else {
      baseParam.methodName = this._schema.method.name;
    }

    const deserializeContext = $page?.id
      ? {
          page: { id: $page.id, type: EComponentType.PAGE },
        }
      : {};

    const processedParams = this._schema.method?.[':params']
      ? deserializeDynamicMap(deserializeContext, { ':params': this._schema.method.params })
      : deserializeDynamicMap(deserializeContext, { params: this._schema.method.params });

    const { staticData, dataBinds = [] } = splitDynamicMap(processedParams);

    this._paramsRef.current = staticData;
    this._dataBinds = generateDataBinds({ $page, $app }, dataBinds, {
      looseError,
      useScopeWidgetData,
      ...options,
      key: `${this.id} method.params`,
    });

    this._action = async (data) => {
      return $w.cloud.callDataSource({
        ...baseParam,
        params: data,
      });
    };

    this._emit = async (trigger, eventData, originalEvent?) =>
      generateEmit($w)(
        trigger,
        generateListnerInstances(
          {
            $app,
            $page,
            actionsMap: (this._context as any).actionsMap,
          },
          this._schema.listeners || [],
        ),
        eventData,
        {},
        originalEvent,
        {},
        {},
        $w.page.id,
        true,
      );

    return this;
  }

  get id(): string {
    return this._schema?.id || '';
  }

  get label(): string {
    return this._schema?.label || '';
  }

  get description(): string {
    return this._schema?.description || '';
  }

  get data() {
    return this._observableValue.data;
  }

  get error() {
    return this._observableValue.error;
  }

  async trigger(additionalScope?, options = {}) {
    this._triggered = true;
    return this._innerTrigger(this._resolveParams(), additionalScope, options);
  }

  reset() {
    this._observableValue.data = null;
    this._observableValue.error = null;
  }

  async _innerTrigger(data, additionalScope?, options = {}) {
    this._currentRequestKey = Date.now();
    const key = this._currentRequestKey;
    try {
      const res = await this._action(data);
      if (key === this._currentRequestKey) {
        this._observableValue.data = res;
        this._observableValue.error = null;
        this._emit(`success`, res);
      }
      return res;
    } catch (e) {
      if (key === this._currentRequestKey) {
        this._observableValue.data = null;
        this._observableValue.error = e;
        this._emit(`fail`, e);
      }
      throw e;
    }
  }

  private _debounceTrigger = debounce(this._innerTrigger, 300);

  destroy() {
    this._disposes.forEach((dispose) => dispose());
  }

  private _resolveParams() {
    /**
     * 这里万一其中某个字段计算失败
     * 好像会阻塞其他字段的计算
     * 从而导致 autorun 没有添加依赖
     */
    return mergeDynamic2StaticData(toJS(this._paramsRef.current), this._dataBinds, {
      $w: this._context.$w,
      // may be additional scope
    })?.params;
  }
}
